#include <ntddk.h>
#include <conio.h>

#define IOCTL_WBYTE CTL_CODE(FILE_DEVICE_UNKNOWN,\
                             0x801,\
                             METHOD_BUFFERED,\
                             FILE_ANY_ACCESS)

#define IOCTL_RBYTE CTL_CODE(FILE_DEVICE_UNKNOWN,\
                             0x802,\
                             METHOD_BUFFERED,\
                             FILE_ANY_ACCESS)

#define NT_DEVICE_NAME L"\\Device\\comdevice"
#define DOS_DEVICE_NAME L"\\DosDevices\\comdevice"

typedef struct _MY_DEVICE_EXTENSION
{
    USHORT Addr;
    UCHAR Data;
} MY_DEVICE_EXTENSION, *PMY_DEVICE_EXTENSION;

VOID
comdevice_Unload (IN PDRIVER_OBJECT DriverObject)
{
   UNICODE_STRING DosDeviceName;
   RtlInitUnicodeString(&DosDeviceName, DOS_DEVICE_NAME);
   IoDeleteSymbolicLink(&DosDeviceName);
   IoDeleteDevice(DriverObject->DeviceObject);
}

NTSTATUS
comdevice_Create(IN PDEVICE_OBJECT DeviceObject,
                 IN PIRP Irp)
{
  Irp->IoStatus.Status = STATUS_SUCCESS;
  IoCompleteRequest(Irp, IO_NO_INCREMENT);
  return STATUS_SUCCESS;
}

NTSTATUS
comdevice_Close(IN PDEVICE_OBJECT DeviceObject,
                IN PIRP Irp)
{
  Irp->IoStatus.Status = STATUS_SUCCESS;
  IoCompleteRequest(Irp, IO_NO_INCREMENT);
  return STATUS_SUCCESS;
}

NTSTATUS
comdevice_IOCTL (IN PDEVICE_OBJECT DeviceObject,
                 IN PIRP Irp)
{
   PMY_DEVICE_EXTENSION dx = (PMY_DEVICE_EXTENSION)DeviceObject->DeviceExtension;
   PIO_STACK_LOCATION pIoStack = IoGetCurrentIrpStackLocation(Irp);
   ULONG ctlCode = pIoStack->Parameters.DeviceIoControl.IoControlCode;

   PUCHAR pBuf = (PUCHAR)Irp->AssociatedIrp.SystemBuffer;
   NTSTATUS status;

switch (ctlCode)
  {
    case IOCTL_WBYTE:
       RtlMoveMemory(&dx->Addr,
                     pBuf,
                     2);
       RtlMoveMemory(&dx->Data,
                     pBuf+2,
                     1);
      _outp(dx->Addr, dx->Data);
    case IOCTL_RBYTE:
       RtlMoveMemory(&dx->Addr,
                     pBuf,
                     2);
       dx->Data = _inp(dx->Addr);
       RtlMoveMemory(pBuf,
                     &dx->Data,
                     1);
      Irp->IoStatus.Information = 1;
  }
Irp->IoStatus.Status = STATUS_SUCCESS;
IoCompleteRequest(Irp, IO_NO_INCREMENT);
return STATUS_SUCCESS;
}

NTSTATUS
DriverEntry (IN PDRIVER_OBJECT DriverObject,
             IN PUNICODE_STRING RegistryPath)
{
  PDEVICE_OBJECT DeviceObject;
  UNICODE_STRING NtDeviceName;
  UNICODE_STRING DosDeviceName;
  NTSTATUS status;

  RtlInitUnicodeString(&NtDeviceName, NT_DEVICE_NAME);
  status = IoCreateDevice(DriverObject,
                          sizeof(MY_DEVICE_EXTENSION),
                          &NtDeviceName,
                          FILE_DEVICE_UNKNOWN,
                          0,
                          FALSE,
                          &DeviceObject);
  if (!NT_SUCCESS(status))
   {
    IoDeleteDevice(DeviceObject);
    return status;
   }

  RtlInitUnicodeString(&DosDeviceName, DOS_DEVICE_NAME);
  status = IoCreateSymbolicLink(&DosDeviceName, &NtDeviceName);
  if (!NT_SUCCESS(status))
   {
    IoDeleteSymbolicLink(&DosDeviceName);
    IoDeleteDevice(DeviceObject);
    return status;
   }

  DeviceObject->Flags &= ~DO_DEVICE_INITIALIZING;
  DeviceObject->Flags |= DO_BUFFERED_IO;

  DriverObject->MajorFunction[IRP_MJ_CREATE] = comdevice_Create;
  DriverObject->MajorFunction[IRP_MJ_CLOSE]  = comdevice_Close;
  DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = comdevice_IOCTL;
  DriverObject->DriverUnload = comdevice_Unload;

  return STATUS_SUCCESS;
}
